"""The cli for the chattie command."""

import click
import os

from inspect import getdoc
from chattie.bot import Bot
from chattie.tricks import helpcmd
from chattie.plugins import get_connectors
from chattie.plugins import get_commands


@click.group()
def chattie():
    """Helpful commands for running a chattie bot."""
    pass


@chattie.command()
def connectors():
    """Show all installed connectors."""
    connectors = get_connectors()
    for c in connectors:
        print(c.name)


@chattie.command()
@click.argument('connector_name')
def doc(connector_name):
    """View the doc for the given connector."""
    connectors = get_connectors()
    for c in connectors:
        if c.name == connector_name:
            print(getdoc(c.load()))


@chattie.command()
def commands():
    """Show all installed commands."""
    import chattie.connectors.term as term
    bot = Bot('Chattie', term, get_commands())
    print(helpcmd(bot, ''))


@chattie.command()
@click.argument('bot_name')
def new(bot_name):
    """Create a new bot."""
    os.mkdir(bot_name)
    # Create the tricks directory
    os.mkdir(os.path.join(bot_name, 'tricks'))
    with open(os.path.join(bot_name, 'tricks', '__init__.py'), 'w') as f:
        f.write("commands = {}")
    # Create the handlers directory
    os.mkdir(os.path.join(bot_name, 'handlers'))
    with open(os.path.join(bot_name, 'handlers', '__init__.py'), 'w') as f:
        f.write("handlers = []")

    with open(os.path.join(bot_name, 'envfile'), 'w') as f:
        f.write("""export BOT_NAME=%s

######################################################################
                    Telegram Connector Config
######################################################################
# Make sure you install python-telegram-bot first with
# pip3 install python-telegram-bot
# export TELEGRAM_API_TOKEN='your token here'

######################################################################
                    Matrix Connector Config
######################################################################
# Make sure you install matrix_client first with
# pip3 install matrix_client
# export MATRIX_URL='https://matrix.org'
# export MATRIX_USERNAME=%s
# export MATRIX_PASSWORD='my_bots_password'
# Comma seperated list of rooms to join
# export MATRIX_ROOMS='#thorin-test:matrix.org'""" % bot_name)

    with open(os.path.join(bot_name, 'README'), 'w') as f:
        f.write("""How to use your new bot!

First you can write new commands or 'tricks' in the tricks folder and
new handlers in the handlers folder. If that doesn't make sense to you
you can view the documentation at:

https://github.com/chasinglogic/Chattie

Now if you want to just run your bot you need to first pick a
connector. To view installed connectors run:

chattie connectors

Once you've picked a connector then set the appropriate environment
variables in envfile then activate it with:

source envfile

Finally:

chattie run --connector connector_name

and you're good to go!
""")


@chattie.command()
@click.option('--name',
              default=os.getenv('BOT_NAME', 'Chattie'),
              help='The name of your bot.')
@click.option('--connector',
              default='terminal_connector',
              help='Which connector to use, see "chattie connectors"')
def run(name, connector):
    """Run the bot.

    By default will run with the first available connector.
    """
    conn = None
    available = get_connectors()
    for c in available:
        if c.name == connector:
            conn = c.load()

    commands = get_commands()
    print('comm', commands)
    bot = Bot(name, conn, commands)
    bot.run()


if __name__ == '__main__':
    chattie()



"""The default commands which come with Chattie."""

import requests
import random

from os import getenv
from inspect import getdoc


def helpcmd(bot, inc_msg):
    """Show the available commands and documentation."""
    try:
        msg = """
I know the following commands:

"""
        for name, fun in bot.commands.items():
            msg += """%s
--------
%s
========

""" % (name, getdoc(fun))
        return msg
    except:
        return 'Sorry, I don\'t know any commands'


def pick(bot, inc_msg):
    """Choose a random item from a list of things."""
    choices = inc_msg[inc_msg.index("pick") + 1:]
    return choices[random.randrange(0, len(choices))]


def the_rules(bot, incoming):
    """Print THE RULES."""
    return """The rules are:
    1. A robot may not injure a human being or, through
       inaction, allow a human being to come to harm.
    2. A robot must obey orders given it by human beings
       except where such orders would conflict with the First Law.
    3. A robot must protect its own existence as long as such
       protection does not conflict with the First or Second Law."""


def yoda(bot, inc_msg):
    """Attempt to translate the given message into yoda-speak."""
    msg = ' '.join(inc_msg[inc_msg.index('yoda') + 1:])
    url = "http://api.funtranslations.com/translate/yoda.json?text=" + msg
    r = requests.get(url)
    if r.status_code == 429:
        return "Sorry I can only translate 5 times per hour"
    try:
        return r.json()['contents']['translated'].replace("  ", " ")
    except:
        print(r.text)
        return "An unexpected error occured"


def hello(bot, incoming):
    """A simple ping style command for testing connectivity."""
    return "Wazzup"


JENKINS_URL = getenv("JENKINS_URL")
JENKINS_API_KEY = getenv("JENKINS_TOKEN")


def build(bot, inc_msg):
    """Run the given build in Jenkins."""
    r = requests.post(JENKINS_URL + "/job/" +
                      ' '.join(inc_msg[inc_msg.index("build") + 1:]) +
                      "/build?token=" + JENKINS_API_KEY)
    if r.status_code > 300:
        print("Error with the request:", r.text)
        return "I couldn't start the build."
    else:
        return "Build successfully started."


WEATHER_API_KEY = getenv("CHATTIE_WEATHER_API_TOKEN")


def _is_zip_code(requested):
    try:
        int(requested)
        return True
    except:
        return False


def _format_resp(jsn):
    return 'Temperature: %s Condition: %s'.format(
        jsn['main']['temp'],
        jsn['weather'][0]['description']
    )


# We make a call to:
# api.openweathermap.org/data/2.5/weather?q={city name}
def _get_weather_name(city):
    r = requests.get("http://api.openweathermap.org/data/2.5/weather?q=" +
                     city + "&APPID=" + WEATHER_API_KEY + "&units=imperial")
    if r.status_code == 404:
        return "I can't find that city"
    if r.status_code > 300:
        return "Some unknown error occurred"
    return _format_resp(r.json())


# api.openweathermap.org/data/2.5/weather?zip={zip code}
def _get_weather_zip(zipcode):
    r = requests.get("http://api.openweathermap.org/data/2.5/weather?zip="
                     + zipcode + "&APPID=" + WEATHER_API_KEY + "&units=imperial")
    if r.status_code == 404:
        return "I can't find that city"
    if r.status_code > 300:
        return "Some unknown error occurred"
    return _format_resp(r.json())


def weather(chattie, incoming):
    """Get the weather for a given city name or zip code."""
    city_or_zip = str(incoming[incoming.index("weather"):])
    if _is_zip_code(city_or_zip):
        return _get_weather_zip(city_or_zip)
    return _get_weather_name(city_or_zip)


commands = {
    'help': helpcmd,
    'pick': pick,
    'the_rules': the_rules,
    'yoda': yoda,
    'hello': hello
}

if WEATHER_API_KEY:
    commands['weather'] = weather.weather

if JENKINS_API_KEY:
    commands['build'] = build



"""A connector for the terminal. Useful for debugging.

It simply opens a REPL that lets you send "messages" to your bot.
"""


class Connector:
    """A connector for the terminal. Useful for debugging."""

    def __init__(self, parser):
        """Set the parser."""
        self.parser = parser

    def listen(self):
        """Listen for messages."""
        print('Make sure to include your bot name in the message.'
              ' Type quit to exit.')

        while True:
            msg = input('>> ')
            if msg == 'quit':
                break
            self.parser(0, msg)

    def send_message(self, room_id, msg):
        """Print Chattie's response."""
        print(msg)



"""The primary Bot class which handles inventory and connections."""

import json
import sys
import os

from os.path import isfile
from os.path import exists


class Bot:
    """Base Bot class, maintains state and parsing commands."""

    inventory = {}

    def __init__(self, name, connector, command_pkgs, handlers=[]):
        """Initialize the bot.

        connector should be a module which contains a class named
        Connector that follows the appropriate interface. See
        chattie.connectors for examples.

        command_pkgs should be a list of packages as returned by
        get_commands() from chattie.plugins. (essentially as returned
        by the entry_points functions)

        A command package needs to have a global dict variable named
        commands which contains a key for each command name and a
        corresponding value which is the function to call for that
        command. The command functions will be called with two
        arguments the first being the current instance of the Bot
        class the second will be an argv like array of the message.

        See the examples directory for commands, connectors, and handlers
        """
        print("Booting systems...")
        self.name = name
        print("Hello my name is " + name + "...")
        self.connector = connector.Connector(self.parse_message)
        if isfile("./inventory.json"):
            print("Loading my inventory from last time...")
            self.__load_inventory()
        self.handlers = handlers
        self.commands = {}
        for pkg in command_pkgs:
            loaded = pkg.load()
            self.commands.update(loaded.commands)

        # Add current directory PYTHONPATH for dynamic imports.
        sys.path.append(os.getcwd())

        # Check if tricks exists and add it if so.
        if exists('./tricks'):
            import tricks
            self.commands.update(tricks.commands)

        # Look for local handlers
        if exists('./handlers'):
            import handlers
            self.handlers += handlers.handlers

    def run(self):
        """Run the bot."""
        print("I am listening for messages...")
        self.connector.listen()

    def get(self, key):
        """Get key from the inventory."""
        return self.inventory[key]

    def set(self, key, value):
        """Save value in the inventory at key."""
        self.inventory[key] = value
        self.__save_inventory()

    def __load_inventory(self):
        """Load the inventory file from the filesystem.

        Potentially destructive function so we attempt to privatize it.
        """
        with open("./inventory.json", "r") as inv:
            self.inventory = json.load(inv)

    def __save_inventory(self):
        """Save the inventory to the file system.

        Potentially destructive function so we attempt to privatize it.
        """
        with open("./inventory.json", "w") as inv:
            json.dump(self.inventory, inv)

    def parse_message(self, room_id, msg):
        """Turn the message into an array and calls the requested command."""
        print("Message received...")
        if self.name.lower() in msg.lower():
            print("Someone is talking to me...")
            split = msg.split(" ")
            print("Parsed: ", split)
            # get the first word after our name as that will be the
            # command always.
            cmd_idx = 1
            for i, w in enumerate(split):
                if w.lower().endswith(self.name.lower()):
                    cmd_idx = i + 1
                    break

            cmd = self.commands.get(split[cmd_idx])
            if cmd is None:
                self.connector.send_message(room_id,
                                            'I don\'t know that trick.')
                return
            reply = cmd(self, split)
            self.connector.send_message(room_id, reply)
            return
        # If no command then pass to handlers
        for h in self.handlers:
            reply = h(self, msg)
            if reply:
                self.send_message(room_id, reply)



"""The reference Connector class that defines the required interface."""


class Connector:
    """The base interface that must be implemented by a backend."""

    def __init__(self, parser):
        """Parser is the parse_message function of the Bot class.

        It should be passed the room_id (whatever form that takes) and
        the plain text of the incoming message from the service.
        """
        self.parser = parser

    def listen(self):
        """Should connect and listen to incoming messages from the backend.

        When an incoming message is parsed should send to self.parser
        (the Bot class' parse_message method)
        """
        pass

    def send_message(self, room_id, msg):
        """Send the msg to room_id.

        msg is always a plain string and room_id is whatever is passed
        to the Bot classes parse_message method as the room_id.
        """
        pass



"""A connector for the Telegram bot API.

This connector only requires the TELEGRAM_API_TOKEN environment
variable is set. You can acquire a token and instructions for making
your bot join rooms from here: https://core.telegram.org/bots
"""

import os

try:
    from telegram.ext import Updater
    from telegram.ext import MessageHandler
except ImportError:
    import sys
    print('You need to pip3 install python-telegram-bot before '
          'using this connector!')
    sys.exit(1)



class Connector:
    """Connector class for the Telegram bot API."""

    # Holds the bots for all given rooms
    bots = {}

    def __init__(self, parser):
        """Will load the api token from $TELEGRAM_API_TOKEN."""
        token = os.getenv('TELEGRAM_API_TOKEN')
        if token is None:
            raise Exception('TELEGRAM_API_TOKEN not set')
        self.updater = Updater(token=token)
        self.dispatcher = self.updater.dispatcher
        self.parser = parser
        self.dispatcher.add_handler(MessageHandler(None, self.parse_incoming))

    def listen(self):
        """Listen for messages."""
        self.updater.start_polling()
        self.updater.idle()

    def send_message(self, room_id, msg):
        """Send a message to the given room_id."""
        self.bots[str(room_id)].sendMessage(chat_id=room_id, text=msg)

    def parse_incoming(self, bot, incoming):
        """Transform incoming telegram info into format Chattie can parse."""
        self.bots[str(incoming.message.chat_id)] = bot
        self.parser(incoming.message.chat_id, incoming.message.text)



"""A Chattie connector for Matrix.

This connector requires that you register an account for the bot
before setting up a connection

Requires the following environment variables:

MATRIX_URL -- The Matrix homeserver, such as https://matrix.org
MATRIX_USERNAME -- The bot username
MATRIX_PASSWORD -- The bot password
MATRIX_ROOMS -- a comma seperated list of rooms for the bot to join
"""

import os

try:
    from matrix_client.client import MatrixRequestError
    from matrix_client.client import MatrixClient
except ImportError:
    import sys
    print('You need to pip3 install matrix_client before '
          'using this connector!')
    sys.exit(1)

MATRIX_URL = os.getenv("MATRIX_URL")
MATRIX_USERNAME = os.getenv("MATRIX_USERNAME")
MATRIX_PASSWORD = os.getenv("MATRIX_PASSWORD")
MATRIX_ROOMS = os.getenv("MATRIX_ROOMS")


class Connector:
    """A matrix connector for the Chattie bot framework."""

    def __init__(self, parser):
        """Connect to the matrix room."""
        self.parser = parser
        self.client = MatrixClient(MATRIX_URL)
        # Try to register if it fails try to log in.
        try:
            self.token = self.client.\
                         register_with_password(username=MATRIX_USERNAME,
                                                password=MATRIX_PASSWORD)
        except MatrixRequestError:
            self.token = self.client.\
                         login_with_password(username=MATRIX_USERNAME,
                                             password=MATRIX_PASSWORD)
        self.rooms = {}
        for room in MATRIX_ROOMS.split(","):
            room_name = room
            if not room.startswith("#"):
                room_name = "!" + room
            r = self.client.join_room(room_name)
            r.add_listener(self.parse_event)
            self.rooms[r.room_id] = r

    def listen(self):
        """Listen for messages on Matrix."""
        self.client.listen_forever()

    def send_message(self, room_id, msg):
        """Send a message to Matrix."""
        self.rooms[room_id].send_text(msg)

    def parse_event(self, room, incoming_event):
        """Transform Matrix incoming_event to text."""
        if incoming_event['type'] != 'm.room.message':
            return
        self.parser(incoming_event['room_id'],
                    incoming_event['content']['body'])



import os
import re
from setuptools import setup
from setuptools import find_packages


chattiefile = os.path.join(os.path.dirname(__file__),
                           'src', 'chattie', '__init__.py')

# Thanks to SQLAlchemy:
# https://github.com/zzzeek/sqlalchemy/blob/master/setup.py#L104
with open(chattiefile) as stream:
    __version__ = re.compile(
        r".*__version__ = '(.*?)'", re.S
    ).match(stream.read()).group(1)


setup(
    name='chattie',
    description='A framework for making bots in Python. Inspired by Hubot',
    version=__version__,
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    author="Mathew Robinson",
    author_email="mrobinson@praelatus.io",
    download_url='https://github.com/chasinglogic/Chattie',
    install_requires=[
        'requests',
        'click'
    ],
    extras_requires={
        'matrix': 'matrix_client',
        'telegram': 'telegram'
    },
    entry_points={
        'console_scripts': [
            'chattie = chattie.cli:chattie'
        ],
        'chattie.plugins.tricks': [
            'default_tricks = chattie.tricks',
        ],
        'chattie.plugins.connectors': [
            'telegram = chattie.connectors.telegram',
            'terminal = chattie.connectors.term',
            'matrix = chattie.connectors.matrix'
        ]
    },
    license='Apache2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ]
)



"""Contains methods for finding commands and connectors."""

import pkg_resources


def get_connectors():
    """Find all connectors available on the system."""
    return [v for v in
            pkg_resources.iter_entry_points('chattie.plugins.connectors')]


def get_commands():
    """Find all commands available on the system."""
    return [v for v in
            pkg_resources.iter_entry_points('chattie.plugins.tricks')]


def get_handlers():
    """Find all handlers available on the system."""
    mods = [v.load() for v in
            pkg_resources.iter_entry_points('chattie.plugins.handlers')]
    handlers = []
    for m in mods:
        handlers += m.handlers
    return handlers
